#include <stdio.h>
#include "hexfile.h"


hexfile * hfOpen(hexfile *hf, const char *hfn)
{
  hf->hfp = fopen(hfn, "r");
  hf->high16 = 0x0000;
  return (hf->hfp) ? hf : NULL;
}

hexfile * hfOpenWrite(hexfile *hf, const char *hfn)
{
  hf->hfp = fopen(hfn, "w");
  hf->high16 = 0x0000;
  hf->eofwritten = 0;
  return (hf->hfp) ? hf : NULL;
}


void hfClose(hexfile *hf)
{
  if (!(hf->eofwritten)) hfWriteEof(hf);
  fclose(hf->hfp);
}


char hexval(const char ch)
{
  char val;

  val = ch - '0';
  if (val > 9) val -= 7;
  return val;
}


int hfReadAsWords(hexfile *hf, hexwords *hw)
{
  char buf[1024];
  unsigned char rtype;
  unsigned char bcount;
  unsigned short baddr;
  int i;
  int w;
  int b;

  if (fgets(buf, sizeof(buf), hf->hfp) == NULL) {
    return -1;
  }

  // Make sure first charater is start code
  if (buf[0] != ':') {
    hw->count = 0;
    return 0;
  }

  // Read the byte count and convert to words
  bcount = (hexval(buf[1]) * 16) + hexval(buf[2]);
  hw->count = bcount / 2;

  // Read the start address and convert to words
  baddr = 0x0000;
  for (i=0; i<=3; i++) {
    baddr = (baddr << 4) + hexval(buf[3+i]);
  }
  hw->start = baddr / 2;

  rtype = (hexval(buf[7]) * 16) + hexval(buf[8]);

  switch (rtype) {
    case 0x00:
      // Read the data into hw->words as words (4 characters each)
      for (w=0; w<(hw->count); w++) {
        b = 9 + (w * 4);
        hw->words[w] = (hexval(buf[b]) * 16) + (hexval(buf[b+1]))
                     + (hexval(buf[b+2]) * (256*16)) + (hexval(buf[b+3]) * 256);
      }
      if (hf->high16 & 0x0001) hw->start = (hw->start | 0x8000);
      hw->high16 = hf->high16 >> 1;
      break;
    case 0x01:
      // Termination record
      hw->count = 255;
      return -1;
      break;
    case 0x04:
      // Extended linear address record
      b = 9;
      hf->high16 = (hexval(buf[b]) * (256*16)) + (hexval(buf[b+1]) * 256)
                 + (hexval(buf[b+2]) * 16) + (hexval(buf[b+3]));
      hw->high16 = hf->high16;
      hw->count = 0;
      // fprintf(stderr, "XLA: high16 = 0x%x\n", hf->high16);
      return 0;
  } // switch
  return hw->count;
}


unsigned char hfCalcChecksum(unsigned char *b, int l)
{
  unsigned int a = 0;
  int i;

  for (i=0; i<l; i++) a += b[i];
  return 0x01 + ~(a & 0xff);
}


void hfWriteHex(hexfile *hf, unsigned char *b, int l)
{
  int i;

  fprintf(hf->hfp, ":");
  for (i=0; i<l; i++) fprintf(hf->hfp, "%02X", b[i]);
  fprintf(hf->hfp, "\n");
}


void hfSetHigh16(hexfile *hf, const unsigned short h16)
{
  unsigned char bytes[64];
  int length = 0;

  bytes[length++] = 0x02;  // Data length
  bytes[length++] = 0x00;
  bytes[length++] = 0x00;
  bytes[length++] = 0x04;  // ELA
  bytes[length++] = (h16 & 0xff00) >> 8;
  bytes[length++] = (h16 & 0x00ff);
  bytes[length] = hfCalcChecksum(bytes, length);
  length++;
  hfWriteHex(hf, bytes, length);
  hf->high16 = h16;
}


int hfWriteWords(hexfile *hf, hexwords *hw)
{
  unsigned char bytes[64];
  int length = 0;
  int i;
  unsigned short high16;

  high16 = (hw->high16 << 1) + (hw->start >> 15);
  if (hf->high16 != high16) hfSetHigh16(hf, high16);
  bytes[length++] = hw->count * 2;
  bytes[length++] = ((hw->start << 1) & 0xff00) >> 8;
  bytes[length++] = ((hw->start << 1) & 0x00ff);
  bytes[length++] = 0x00;  // Data record
  for (i=0; i< hw->count; i++) {
    bytes[length++] = (hw->words[i] & 0x00ff);
    bytes[length++] = (hw->words[i] & 0xff00) >> 8;
  }
  bytes[length] = hfCalcChecksum(bytes, length);
  length++;
  hfWriteHex(hf, bytes, length);
  return length;
}


void hfWriteEof(hexfile *hf)
{
  fprintf(hf->hfp, ":00000001FF\n");
  hf->eofwritten = 1;
}

