#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include "hexfile.h"
#include "memmap.h"

#define HEXFILE_LINESIZE	(8)


mmhandle mmInit(const int pmsize, const int cmsize, const int eesize)
{
  mmhandle mh = (mmhandle)malloc(sizeof(memmapinfo));
  mh->psize = pmsize;
  mh->csize = cmsize;
  mh->esize = eesize;
  mh->pm = (short *)malloc(pmsize * sizeof(short));
  mh->cm = (short *)malloc(cmsize * sizeof(short));
  mh->em = (short *)malloc(eesize * sizeof(short));

  // Allocate bitmaps for locations with valid data
  mh->pmap = (unsigned int *)malloc(((pmsize / 32) + 1) * sizeof(unsigned int));
  mh->cmap = (unsigned int *)malloc(((cmsize / 32) + 1) * sizeof(unsigned int));
  mh->emap = (unsigned int *)malloc(((eesize / 32) + 1) * sizeof(unsigned int));

  // Set all memory to erased state (on PICs, that's all 1's)
  memset(mh->pm, 0xff, mh->psize * sizeof(short));
  memset(mh->cm, 0xff, mh->csize * sizeof(short));
  memset(mh->em, 0xff, mh->esize * sizeof(short));

  // Zero out bitmaps
  memset(mh->pmap, 0x00, ((mh->psize / 32) + 1) * sizeof(unsigned int));
  memset(mh->cmap, 0x00, ((mh->csize / 32) + 1) * sizeof(unsigned int));
  memset(mh->emap, 0x00, ((mh->esize / 32) + 1) * sizeof(unsigned int));

  // Finally, return the memmap handle
  return mh;
}


void mmFree(mmhandle h)
{
  free(h->emap);
  free(h->cmap);
  free(h->pmap);
  free(h->em);
  free(h->cm);
  free(h->pm);
  free(h);
}


// NB: Since the map size is not passed in, bounds checking
// *MUST* be done by the caller. This might change in the
// future, requiring additional argument of map size.

int mmBitSet(unsigned int *bm, const unsigned int a)
{
  unsigned int loc = a / 32;
  unsigned int bitnum = a % 32;
  unsigned int orval = (1 << bitnum);

  bm[loc] = bm[loc] | orval;
  return 0;
}


int mmBitClear(unsigned int *bm, const unsigned int a)
{
  unsigned int loc = a / 32;
  unsigned int bitnum = a % 32;
  unsigned int andval = (1 << bitnum) ^ 0xffffffff;

  bm[loc] = bm[loc] & andval;
  return 0;
}


int mmBitTest(unsigned int *bm, const unsigned int a)
{
  unsigned int loc = a / 32;
  unsigned int bitnum = a % 32;
  unsigned int andval = (1 << bitnum);

  return (bm[loc] & andval) ? 1 : 0;
}


int mmWrite(mmhandle h, const unsigned int pc, const unsigned short w)
{
  int validwrite = -1;
  unsigned int relpc;

  if (pc < h->psize) {
    // Write to program memory
    // fprintf(stderr, "P");
    validwrite = pc;
    relpc = pc;
    (h->pm)[relpc] = w;
    mmBitSet(h->pmap, relpc);
  }

  if ((pc >= 0x8000) && (pc < (0x8000 + h->csize))) {
    // Write to configuration memory
    // fprintf(stderr, "C");
    validwrite = pc;
    relpc = pc - 0x8000;
    (h->cm)[relpc] = w;
    mmBitSet(h->cmap, relpc);
  }

  if ((pc >= 0xf000) && (pc < (0xf000 + h->esize))) {
    // Write to EEPROM memory
    // fprintf(stderr, "E");
    validwrite = pc;
    relpc = pc - 0xf000;
    (h->em)[relpc] = w;
    mmBitSet(h->emap, relpc);
  }
  if (validwrite < 0) {
    fprintf(stderr, "mmRead(): Attempt to write invalid address %04x\n", pc);
  }
  return validwrite;
}


unsigned short mmRead(mmhandle h, const unsigned int pc)
{
  unsigned int relpc;

  if (pc < h->psize) {
    // Read from program memory
    relpc = pc;
    return (h->pm)[relpc] & 0xffff;
  }

  if ((pc >= 0x8000) && (pc < (0x8000 + h->csize))) {
    // Read from configuration memory
    relpc = pc - 0x8000;
    return (h->cm)[relpc] & 0xffff;
  }

  if ((pc >= 0xf000) && (pc < (0xf000 + h->esize))) {
    // Read from EEPROM memory
    relpc = pc - 0xf000;
    return (h->em)[relpc] & 0xffff;
  }

  fprintf(stderr, "mmRead(): Attempt to read invalid address %04x\n", pc);
  return 0xffff;
}


/*
 * RETURNS	1 if location has data, 0 if not, -1 if out of range.
 */

int mmHasData(const mmhandle h, const unsigned int pc)
{
  if (pc < h->psize) {
    // Program memory
    return mmBitTest(h->pmap, pc);
  }

  if ((pc >= 0x8000) && (pc < (0x8000 + h->csize))) {
    // Config memory
    return mmBitTest(h->cmap, pc - 0x8000);
  }

  if ((pc >= 0xf000) && (pc < (0xf000 + h->esize))) {
    // EEPROM memory
    return mmBitTest(h->emap, pc - 0xf000);
  }
  return -1;
}


/*
 * PARAMETERS	'bsize' must be a power of 2 (2,4,8,16,...) or bad
 *		things will probably happen.
 * RETURNS	1 if location has data, 0 if not, -1 if out of range.
 */

int mmBlockHasData(const mmhandle h, const unsigned int pc, const unsigned int bsize)
{
  unsigned int *map = NULL;
  unsigned int relpc;
  unsigned int basepc;
  unsigned int i;

  if (pc < h->psize) {
    map = h->pmap;
    relpc = pc;
  }

  if ((pc >= 0x8000) && (pc < (0x8000 + h->csize))) {
    map = h->cmap;
    relpc = pc - 0x8000;
  }

  if ((pc >= 0xf000) && (pc < (0xf000 + h->esize))) {
    map = h->emap;
    relpc = pc - 0xf000;
  }

  if (map == NULL) return -1;

  // Create a mask from the block size to ensure basepc is
  // on a 'bsize' byte boundary.
  //
  i = (bsize - 1) ^ 0xffffffff;
  basepc = relpc & i;

  for (i=0; i<bsize; i++) {
    if (mmBitTest(map, basepc+i)) return 1;
  }
  return 0;
}


void mmDataSummary(mmhandle h)
{
  int count;
  int i;

  count = 0;
  for (i=0; i < h->psize; i++) {
    if (mmBitTest(h->pmap, i)) count++;
  }
  fprintf(stderr, "Valid PM bytes: %d\n", count);

  count = 0;
  for (i=0; i < h->csize; i++) {
    if (mmBitTest(h->cmap, i)) {
      count++;
      // fprintf(stderr, "%04x: %04x\n", i, (h->cm)[i]);
    } else {
      // fprintf(stderr, "%04x: -\n", i);
    }
  }
  fprintf(stderr, "Valid CM bytes: %d\n", count);

  count = 0;
  for (i=0; i < h->esize; i++) {
    if (mmBitTest(h->emap, i)) count++;
  }
  fprintf(stderr, "Valid EE bytes: %d\n", count);
}



mmhandle mmLoadHexfile(mmhandle h, const char * hfn)
{
  hexfile hf;
  hexwords hw;
  int pc;
  int i;

  if (h == NULL) return NULL;

  if (!hfOpen(&hf, hfn)) {
    fprintf(stderr, "Error opening hexfile: %s\n", hfn);
    return NULL;
  }

  while (hfReadAsWords(&hf, &hw) >= 0) {
    if (hw.count > 0) {
      pc = hw.start;
      for (i=0; i<hw.count; i++) {
        mmWrite(h, pc++, hw.words[i] & 0xffff);
      }
    }  // if (count > 0)
  }

  hfClose(&hf);
  mmDataSummary(h);
  return h;
}


mmhandle mmWriteHexfile(mmhandle h, const char * hfn)
{
  hexfile hf;
  hexwords hw;
  int pc;
  int i;
  int havedata;

  if (h == NULL) return NULL;

  if (!hfOpenWrite(&hf, hfn)) {
    fprintf(stderr, "Error opening hexfile: %s\n", hfn);
    return NULL;
  }

  // Write Program Memory
  for (pc=0; pc<(h->psize); pc += HEXFILE_LINESIZE) {
    if (mmBlockHasData(h, pc, HEXFILE_LINESIZE)) {
      // Write this block out
      hw.high16 = ((pc) & 0xffff0000) >> 16;
      hw.start = ((pc) & 0x0000ffff);
      hw.count = HEXFILE_LINESIZE;
      for (i=0; i<HEXFILE_LINESIZE; i++) {
        hw.words[i] = mmRead(h, pc+i);
      }
      hfWriteWords(&hf, &hw);
    }
  }

  // Write Config Memory
  havedata = 0;
  for (pc=0x8000; pc<(0x8000+(h->csize)); pc++) {
    if (mmHasData(h, pc)) {
      // Have data here
      if (havedata) {
        hw.words[hw.count] = mmRead(h, pc);
        hw.count++;
      } else {
        hw.high16 = ((pc) & 0xffff0000) >> 16;
        hw.start = ((pc) & 0x0000ffff);
        hw.count = 1;
        hw.words[0] = mmRead(h, pc);
        havedata = 1;
      }
    } else {
      // Data gap
      if (havedata) {
        // No new data, but something pending to be written
        hfWriteWords(&hf, &hw);
        havedata = 0;
      } else {
        // No data + nothing to be written = do nothing
        havedata = 0;
      }
    }
  }

  // Write EEPROM memory
  for (pc=0xf000; pc<(0xf000+(h->esize)); pc += HEXFILE_LINESIZE) {
    if (mmBlockHasData(h, pc, HEXFILE_LINESIZE)) {
      // Write this block out
      hw.high16 = ((pc) & 0xffff0000) >> 16;
      hw.start = ((pc) & 0x0000ffff);
      hw.count = HEXFILE_LINESIZE;
      for (i=0; i<HEXFILE_LINESIZE; i++) {
        hw.words[i] = mmRead(h, pc+i);
      }
      hfWriteWords(&hf, &hw);
    }
  }

  hfClose(&hf);
  mmDataSummary(h);
  return h;
}

