#include <ctype.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include "memmap.h"
#include "symtab.h"
#include "util.h"
#include "eval.h"




int lc;
int pass;
FILE *lstp;

extern mmhandle chipmem;

#define OC_FORWARD	(1 << 0)	/* Opcode had a forward reference */
					/* in pass 1                      */

struct oc_info {
  struct oc_info *next;
  int addr;
  unsigned int flags;
};

#define INFOTAB_SIZE	32

struct oc_info *oc_infotab[INFOTAB_SIZE];


void ocitInit(void)
{
  int i;

  for(i=0; i<INFOTAB_SIZE; i++) {
    oc_infotab[i] = NULL;
  }
}


void ocitDestroy(void)
{
  int i;
  struct oc_info *op;

  for (i=0; i<INFOTAB_SIZE; i++) {
    while (oc_infotab[i] != NULL) {
      op = (oc_infotab[i])->next;
      free(oc_infotab[i]);
      oc_infotab[i] = op;
    }
  }
}


struct oc_info *ocitFindEntry(const int loc)
{
  int i;
  struct oc_info *op;

  i = (loc % INFOTAB_SIZE);
  op = oc_infotab[i];
  while (op) {
    if (op->addr == loc) return op;
    op = op->next;
  }
  return NULL;
}


void ocitSetFlag(int loc, unsigned int f)
{
  int i;
  struct oc_info *oent;

  // fprintf(stderr, "%04x ocitSetFlag: Setting flags %04x at address %04x\n", lc, f, loc);
  if ((oent = ocitFindEntry(loc)) != NULL) {
    oent->flags |= f;
  } else {
    i = (loc % INFOTAB_SIZE);
    oent = (struct oc_info *)malloc(sizeof(struct oc_info));
    oent->addr = loc;
    oent->flags = f;
    oent->next = oc_infotab[i];
    oc_infotab[i] = oent;
  }
}


/*
 *  Checks if flag(s) 'f' is/are set for location 'loc'. Returns
 *  bitwise AND of flags and 'f'. Return value is 'f' if all flags
 *  set, >0 if any flag is set, and 0 if no flags are set.
 */
unsigned int ocitCheckFlag(const int loc, unsigned int f)
{
  int i;
  struct oc_info *op;

  i = (loc % INFOTAB_SIZE);
  op = oc_infotab[i];
  while (op) {
    if (op->addr == loc) return (op->flags & f);
    op = op->next;
  }
  return 0;
}



void split_operands(char *op, char **op1, char **op2)
{
  char *temp_op;
  char quoting = '\0';

  // fprintf(stderr, "%04x DEBUG: Entering split_operands: '%s'\n", lc, op);
  *op1 = op;
  temp_op = op;
  while (*temp_op) {
    if (quoting && (*temp_op == quoting)) {
      quoting = '\0';
      temp_op++;
      continue;
    }
    if ((*temp_op == '\'') || (*temp_op == '"')) {
      quoting = *temp_op;
      temp_op++;
      continue;
    }
    if (*temp_op == ',') {
      *temp_op = '\0';
      temp_op++;
      while (*temp_op != '\0') {
        if ((*temp_op != ' ') && (*temp_op != '\t')) break;
        temp_op++;
      }
      *op2 = temp_op;
      return;
    }
    temp_op++;
  }
  fprintf(stderr, "     Never found operand 2\n");
  *op2 = NULL;
}


int parse_fb(const char *op, unsigned short *f, unsigned short *b)
{
  char buf[128];
  char *fstr;
  char *bstr;

  strncpy(buf,op, sizeof(buf)-1);
  split_operands(buf, &fstr, &bstr);
  // fprintf(stderr, "DEBUG: parse_fb: f = '%s', b = '%s'\n", fstr, bstr);

  if ((fstr[0] == '\0') || (bstr[0] == '\0')) return -1;
  *f = symResolveLiteral(fstr) & 0x007f;
  *b = symResolveLiteral(bstr) & 0x0007;
  return *f;
}



int parse_fd(const char *op, unsigned short *f, unsigned short *d)
{
  char buf[128];
  char *fstr;
  char *dstr;

  strncpy(buf,op, sizeof(buf)-1);
  split_operands(buf, &fstr, &dstr);
  // fprintf(stderr, "DEBUG: parse_fd: f = '%s', d = '%s'\n", fstr, dstr);

  if ((fstr[0] == '\0') || (dstr[0] == '\0')) return -1;
  *f = symResolveLiteral(fstr) & 0x007f;
  lowercase(dstr);
  *d = 0xffff;
  if ((dstr[0] == '0') || (dstr[0] == 'w')) *d = 0;
  if ((dstr[0] == '1') || (dstr[0] == 'f')) *d = 1;
  if (*d == 0xffff) {
    message(MSG_WARNING, "Invalid destination '%s'. Assuming W", dstr);
    *d = 0;
    return -1;
  }
  return *f;
}


/***** Byte-Oriented File Register Operations *****/

void do_addwf(const char *op)
{
  unsigned short f;
  unsigned short d;

  if (parse_fd(op, &f, &d) == -1) {
    fprintf(stderr, "ERROR: parse_fb failed to parse operands.\n");
    exit(1);
  }

  if (pass) mmWrite(chipmem, lc, 0x0700 | (d << 7) | f);
  lc++;
}


void do_addwfc(const char *op)
{
  unsigned short f;
  unsigned short d;

  if (parse_fd(op, &f, &d) == -1) {
    fprintf(stderr, "ERROR: parse_fb failed to parse operands.\n");
    exit(1);
  }

  if (pass) mmWrite(chipmem, lc, 0x3d00 | (d << 7) | f);
  lc++;
}


void do_andwf(const char *op)
{
  unsigned short f;
  unsigned short d;

  if (parse_fd(op, &f, &d) == -1) {
    fprintf(stderr, "ERROR: parse_fb failed to parse operands.\n");
    exit(1);
  }

  if (pass) mmWrite(chipmem, lc, 0x0500 | (d << 7) | f);
  lc++;
}


void do_asrf(const char *op)
{
  unsigned short f;
  unsigned short d;

  if (parse_fd(op, &f, &d) == -1) {
    fprintf(stderr, "ERROR: parse_fb failed to parse operands.\n");
    exit(1);
  }

  if (pass) mmWrite(chipmem, lc, 0x3700 | (d << 7) | f);
  lc++;
}


void do_lslf(const char *op)
{
  unsigned short f;
  unsigned short d;

  if (parse_fd(op, &f, &d) == -1) {
    fprintf(stderr, "ERROR: parse_fb failed to parse operands.\n");
    exit(1);
  }

  if (pass) mmWrite(chipmem, lc, 0x3500 | (d << 7) | f);
  lc++;
}


void do_lsrf(const char *op)
{
  unsigned short f;
  unsigned short d;

  if (parse_fd(op, &f, &d) == -1) {
    fprintf(stderr, "ERROR: parse_fb failed to parse operands.\n");
    exit(1);
  }

  if (pass) mmWrite(chipmem, lc, 0x3600 | (d << 7) | f);
  lc++;
}


void do_clrf(const char *op)
{
  if (pass) mmWrite(chipmem, lc, 0x0180 | (symResolveLiteral(op) & 0x7f));
  lc++;
}


void do_clrw(const char *op)
{
  if (pass) mmWrite(chipmem, lc, 0x0100);
  lc++;
}


void do_comf(const char *op)
{
  unsigned short f;
  unsigned short d;

  if (parse_fd(op, &f, &d) == -1) {
    fprintf(stderr, "ERROR: parse_fb failed to parse operands.\n");
    exit(1);
  }

  if (pass) mmWrite(chipmem, lc, 0x0900 | (d << 7) | f);
  lc++;
}


void do_decf(const char *op)
{
  unsigned short f;
  unsigned short d;

  if (parse_fd(op, &f, &d) == -1) {
    fprintf(stderr, "ERROR: parse_fb failed to parse operands.\n");
    exit(1);
  }

  if (pass) mmWrite(chipmem, lc, 0x0300 | (d << 7) | f);
  lc++;
}


void do_incf(const char *op)
{
  unsigned short f;
  unsigned short d;

  if (parse_fd(op, &f, &d) == -1) {
    fprintf(stderr, "ERROR: parse_fb failed to parse operands.\n");
    exit(1);
  }

  if (pass) mmWrite(chipmem, lc, 0x0a00 | (d << 7) | f);
  lc++;
}


void do_iorwf(const char *op)
{
  unsigned short f;
  unsigned short d;

  if (parse_fd(op, &f, &d) == -1) {
    fprintf(stderr, "ERROR: parse_fb failed to parse operands.\n");
    exit(1);
  }

  if (pass) mmWrite(chipmem, lc, 0x0400 | (d << 7) | f);
  lc++;
}


void do_movf(const char *op)
{
  unsigned short f;
  unsigned short d;

  if (parse_fd(op, &f, &d) == -1) {
    fprintf(stderr, "ERROR: parse_fb failed to parse operands.\n");
    exit(1);
  }

  if (pass) mmWrite(chipmem, lc, 0x0800 | (d << 7) | f);
  lc++;
}


void do_movwf(const char *op)
{
  if (pass) mmWrite(chipmem, lc, 0x0080 | (symResolveLiteral(op) & 0x7f));
  lc++;
}


void do_rlf(const char *op)
{
  unsigned short f;
  unsigned short d;

  if (parse_fd(op, &f, &d) == -1) {
    fprintf(stderr, "ERROR: parse_fb failed to parse operands.\n");
    exit(1);
  }

  if (pass) mmWrite(chipmem, lc, 0x0d00 | (d << 7) | f);
  lc++;
}


void do_rrf(const char *op)
{
  unsigned short f;
  unsigned short d;

  if (parse_fd(op, &f, &d) == -1) {
    fprintf(stderr, "ERROR: parse_fb failed to parse operands.\n");
    exit(1);
  }

  if (pass) mmWrite(chipmem, lc, 0x0c00 | (d << 7) | f);
  lc++;
}


void do_subwf(const char *op)
{
  unsigned short f;
  unsigned short d;

  if (parse_fd(op, &f, &d) == -1) {
    fprintf(stderr, "ERROR: parse_fb failed to parse operands.\n");
    exit(1);
  }

  if (pass) mmWrite(chipmem, lc, 0x0200 | (d << 7) | f);
  lc++;
}


void do_subwfb(const char *op)
{
  unsigned short f;
  unsigned short d;

  if (parse_fd(op, &f, &d) == -1) {
    fprintf(stderr, "ERROR: parse_fb failed to parse operands.\n");
    exit(1);
  }

  if (pass) mmWrite(chipmem, lc, 0x3b00 | (d << 7) | f);
  lc++;
}


void do_swapf(const char *op)
{
  unsigned short f;
  unsigned short d;

  if (parse_fd(op, &f, &d) == -1) {
    fprintf(stderr, "ERROR: parse_fb failed to parse operands.\n");
    exit(1);
  }

  if (pass) mmWrite(chipmem, lc, 0x0e00 | (d << 7) | f);
  lc++;
}


void do_xorwf(const char *op)
{
  unsigned short f;
  unsigned short d;

  if (parse_fd(op, &f, &d) == -1) {
    fprintf(stderr, "ERROR: parse_fb failed to parse operands.\n");
    exit(1);
  }

  if (pass) mmWrite(chipmem, lc, 0x0600 | (d << 7) | f);
  lc++;
}



/***** Byte Oriented Skip Operations *****/

void do_decfsz(const char *op)
{
  unsigned short f;
  unsigned short d;

  if (parse_fd(op, &f, &d) == -1) {
    fprintf(stderr, "ERROR: parse_fb failed to parse operands.\n");
    exit(1);
  }

  if (pass) mmWrite(chipmem, lc, 0x0b00 | (d << 7) | f);
  lc++;
}


void do_incfsz(const char *op)
{
  unsigned short f;
  unsigned short d;

  if (parse_fd(op, &f, &d) == -1) {
    fprintf(stderr, "ERROR: parse_fb failed to parse operands.\n");
    exit(1);
  }

  if (pass) mmWrite(chipmem, lc, 0x0f00 | (d << 7) | f);
  lc++;
}



/***** Bit-Oriented File Register Operations *****/

void do_bcf(const char *op)
{
  unsigned short f;
  unsigned short b;

  if (parse_fb(op, &f, &b) == -1) {
    fprintf(stderr, "ERROR: parse_fb failed to parse operands.\n");
    exit(1);
  }

  if (pass) mmWrite(chipmem, lc, 0x1000 | (b << 7) | f);
  lc++;
}


void do_bsf(const char *op)
{
  unsigned short f;
  unsigned short b;

  if (parse_fb(op, &f, &b) == -1) {
    fprintf(stderr, "ERROR: parse_fb failed to parse operands.\n");
    exit(1);
  }

  if (pass) mmWrite(chipmem, lc, 0x1400 | (b << 7) | f);
  lc++;
}



/***** Bit-Oriented Skip Operations *****/

void do_btfsc(const char *op)
{
  unsigned short f;
  unsigned short b;

  if (parse_fb(op, &f, &b) == -1) {
    fprintf(stderr, "ERROR: parse_fb failed to parse operands.\n");
    exit(1);
  }

  if (pass) mmWrite(chipmem, lc, 0x1800 | (b << 7) | f);
  lc++;
}


void do_btfss(const char *op)
{
  unsigned short f;
  unsigned short b;

  if (parse_fb(op, &f, &b) == -1) {
    fprintf(stderr, "ERROR: parse_fb failed to parse operands.\n");
    exit(1);
  }

  if (pass) mmWrite(chipmem, lc, 0x1c00 | (b << 7) | f);
  lc++;
}



/***** Literal Operations *****/

void do_addlw(const char *op)
{
  if (pass) mmWrite(chipmem, lc, 0x3e00 | (symResolveLiteral(op) & 0x00ff));
  lc++;
}


void do_andlw(const char *op)
{
  if (pass) mmWrite(chipmem, lc, 0x3900 | (symResolveLiteral(op) & 0x00ff));
  lc++;
}


void do_iorlw(const char *op)
{
  if (pass) mmWrite(chipmem, lc, 0x3800 | (symResolveLiteral(op) & 0x00ff));
  lc++;
}


void do_movlb(const char *op)
{
  if (pass) mmWrite(chipmem, lc, 0x0020 | (symResolveLiteral(op) & 0x001f));
  lc++;
}


void do_movlp(const char *op)
{
  if (pass) mmWrite(chipmem, lc, 0x3180 | (symResolveLiteral(op) & 0x007f));
  lc++;
}


void do_movlw(const char *op)
{
  if (pass) mmWrite(chipmem, lc, 0x3000 | (symResolveLiteral(op) & 0x00ff));
  lc++;
}


void do_sublw(const char *op)
{
  if (pass) mmWrite(chipmem, lc, 0x3c00 | (symResolveLiteral(op) & 0x00ff));
  lc++;
}


void do_xorlw(const char *op)
{
  if (pass) mmWrite(chipmem, lc, 0x3a00 | (symResolveLiteral(op) & 0x00ff));
  lc++;
}



/***** Control Operations *****/

void
do_bra(const char *op)
{
  int dst;

  dst = symResolveAddr(op);
  if (dst == EVAL_FAIL) {
    if (!pass) {
      message(MSG_INFO, "%04x: Forward label", lc);
      ocitSetFlag(lc, OC_FORWARD);
    }
    if (pass) {
      message(MSG_ERROR, "%04x: Unresolved symbol: %s", lc, op);
      exit(1);
    }
  }
  if (pass) {
    dst = (dst - (lc + 1));

    if ((dst < -256) || (dst > 255)) {
      message(MSG_WARNING, "bra operand (%d) out of range. Code will be wrong.", dst);
    }

    mmWrite(chipmem, lc, 0x3200 | (shorten2s(dst, 9)));
  }
  lc++;
}


void
do_brw(const char *op)
{
  if (pass) mmWrite(chipmem, lc, 0x000b);
  lc++;
}


void
do_call_near(const char *op)
{
  int dst;

  dst = symResolveAddr(op);
  if (dst == EVAL_FAIL) {
    if (!pass) {
      message(MSG_NOTICE, "Forward label; near call forced");
      ocitSetFlag(lc, OC_FORWARD);
    }
    if (pass) {
      message(MSG_ERROR, "Unresolved symbol: %s\n", op);
      exit(1);
    }
  }
  if ((dst == EVAL_FAIL) || ((dst & 0xf800) != (lc & 0xf800))) {
    if (pass) message(MSG_WARNING, "call_near: long call needed, but near forced. Code will be wrong.");
  }
  if (pass) mmWrite(chipmem, lc, 0x2000 | (dst & 0x07ff));
  lc++;
}


void
do_call_far(const char *op)
{
  int dst;

  dst = symResolveAddr(op);
  if (!pass && (dst == -1)) {
    fprintf(stderr, "%04x NOTICE: Foward label; assuming/forcing long call\n", lc);
    ocitSetFlag(lc, OC_FORWARD);
  }

  if (pass && (dst == -1)) {
    fprintf(stderr, "%04x ERROR: Unresolved symbol: %s\n", lc, op);
    exit(1);
  }

  fprintf(stderr, "%04x INFO: call_far: long call forced.\n", lc);
  if (pass) mmWrite(chipmem, lc, 0x3180 | (dst & 0xff00) >> 8);
  lc++;
  if (pass && mmHasData(chipmem, lc))
      message(MSG_WARNING, "Overwriting previously written location");
  if (pass) mmWrite(chipmem, lc, 0x2000 | (dst & 0x07ff));
  lc++;
}


void
do_call(const char *op)
{
  int dst;

  dst = symResolveAddr(op);
  if (dst == EVAL_FAIL) {
    if (!pass) {
      message(MSG_NOTICE, "Foward label; assuming/forcing long call");
      ocitSetFlag(lc, OC_FORWARD);
    }

    if (pass) {
      message(MSG_ERROR, "Unresolved symbol: %s", op);
      exit(1);
    }
  }

  if ((dst == EVAL_FAIL) || ocitCheckFlag(lc, OC_FORWARD) || ((dst & 0xf800) != (lc & 0xf800))) {
    fprintf(stderr, "%04x INFO: call: long call.\n", lc);
    if (pass) mmWrite(chipmem, lc, 0x3180 | (dst & 0xff00) >> 8);
    lc++;
    if (pass && mmHasData(chipmem, lc))
        message(MSG_WARNING, "Overwriting previously written location");
  }
  if (pass) mmWrite(chipmem, lc, 0x2000 | (dst & 0x07ff));
  lc++;
}


void
do_callw(const char *op)
{
  if (pass) mmWrite(chipmem, lc, 0x000a);
  lc++;
}


void
do_goto_near(const char *op)
{
  int dst;

  dst = symResolveAddr(op);
  if (dst == EVAL_FAIL) {
    if (!pass) {
      message(MSG_NOTICE, "Forward label; near goto forced");
      ocitSetFlag(lc, OC_FORWARD);
    }
    if (pass) {
      message(MSG_ERROR, "Unresolved symbol: %s", op);
      exit(1);
    }
  }

  if ((dst == EVAL_FAIL) || ((dst & 0xf800) != (lc & 0xf800))) {
    if (pass) message(MSG_WARNING, "goto_near: long goto required; code will be wrong.");
  }
  if (pass) mmWrite(chipmem, lc, 0x2800 | (dst & 0x07ff));
  lc++;
}


void
do_goto_far(const char *op)
{
  int dst;

  dst = symResolveAddr(op);
  if (dst == EVAL_FAIL) {
    if (!pass) {
      message(MSG_INFO, "Foward label; long call forced");
      ocitSetFlag(lc, OC_FORWARD);
    }

    if (pass) {
      message(MSG_ERROR, "Unresolved symbol: %s", op);
      exit(1);
    }
  }

  message(MSG_INFO, "goto_far: long goto forced.");
  if (pass) mmWrite(chipmem, lc, 0x3180 | (dst & 0xff00) >> 8);
  lc++;
  if (pass && mmHasData(chipmem, lc))
      message(MSG_WARNING, "Overwriting previously written location");
  if (pass) mmWrite(chipmem, lc, 0x2800 | (dst & 0x07ff));
  lc++;
}


void
do_goto(const char *op)
{
  int dst;

  dst = symResolveAddr(op);
  if (dst == EVAL_FAIL) {
    if (!pass) {
      message(MSG_NOTICE, "Forward label; assuming/forcing long goto");
      ocitSetFlag(lc, OC_FORWARD);
    }

    if (pass) {
      message(MSG_ERROR, "Unresolved symbol: %s", op);
      exit(1);
    }
  }

  if ((dst == -1) || ocitCheckFlag(lc, OC_FORWARD) || ((dst & 0xf800) != (lc & 0xf800))) {
    message(MSG_INFO, "goto: long goto.");
    if (pass) mmWrite(chipmem, lc, 0x3180 | (dst & 0xff00) >> 8);
    lc++;
    if (pass && mmHasData(chipmem, lc))
        message(MSG_WARNING, "Overwriting previously written location");
  }
  if (pass) mmWrite(chipmem, lc, 0x2800 | (dst & 0x07ff));
  lc++;
}


void
do_retfie(const char *op)
{
  if (pass) mmWrite(chipmem, lc, 0x0009);
  lc++;
}


void
do_retlw(const char *op)
{
  if (pass) mmWrite(chipmem, lc, 0x3400 | (symResolveLiteral(op) & 0x00ff));
  lc++;
}


void
do_return(const char *op)
{
  if (pass) mmWrite(chipmem, lc, 0x0008);
  lc++;
}



/***** Inherent Operations *****/

void do_clrwdt(const char *op)
{
  if (pass) mmWrite(chipmem, lc, 0x0064);
  lc++;
}

void do_nop(const char *op)
{
  if (pass) mmWrite(chipmem, lc, 0x0000);
  lc++;
}

void do_option(const char *op)
{
  if (pass) mmWrite(chipmem, lc, 0x0062);
  lc++;
}

void do_reset(const char *op)
{
  if (pass) mmWrite(chipmem, lc, 0x0001);
  lc++;
}

void do_sleep(const char *op)
{
  if (pass) mmWrite(chipmem, lc, 0x0063);
  lc++;
}

void do_tris(const char *op)
{
  int f = 0;

  if ((*op >= '5') && (*op <= '7')) {
    f = *op - '0';
  } else if ((toupper(*op) >= 'A') && (toupper(*op) <= 'C')) {
    f = 5 + (*op - 'A');
  } else {
    fprintf(stderr, "ERROR: Opcode TRIS has invalid operand '%s'\n", op);
    exit(1);
  }
  if (pass) mmWrite(chipmem, lc, 0x0060 | f);
  lc++;
}


/***** "C Optimized" opcodes *****/

void do_addfsr0(const char *op)
{
  if (pass) mmWrite(chipmem, lc, 0x3100 | (shorten2s(symResolveLiteral(op), 6)));
  lc++;
}


void do_addfsr1(const char *op)
{
  if (pass) mmWrite(chipmem, lc, 0x3140 | (shorten2s(symResolveLiteral(op), 6)));
  lc++;
}


void do_moviw(const char *op)
{
  int m = -1;
  int rp = 0;
  int r;
  int k;
  char *ch;

  if ((ch = strchr(op, '['))) {
    r = (ch[4] - '0') << 6;
    sscanf(op, "%d", &k);
    
    if (pass) mmWrite(chipmem, lc, 0x3f00 | r | (shorten2s(k, 6) & 0x003f));
    lc++;

    return;
  }

  if (strlen(op) < 6) {
    message(MSG_ERROR, "moviw operand syntax error.");
    exit(1);
  }
  if (op[1] == '+') { m = 0; rp = 5; }
  if (op[1] == '-') { m = 1; rp = 5; }
  if (op[4] == '+') { m = 2; rp = 3; }
  if (op[4] == '-') { m = 3; rp = 3; }
  if ((m < 0) || (rp < 3)) {
    message(MSG_ERROR, "moviw operand syntax error.");
    exit(1);
  }

  r = (op[rp] - '0') << 2;

  if (pass) mmWrite(chipmem, lc, 0x0010 | r | m);
  lc++;
}


void do_movwi(const char *op)
{
  int m = -1;
  int rp = 0;
  int r;
  int k;
  char *ch;

  if ((ch = strchr(op, '['))) {
    r = (ch[4] - '0') << 6;
    sscanf(op, "%d", &k);
    
    if (pass) mmWrite(chipmem, lc, 0x3f80 | r | (shorten2s(k, 6) & 0x003f));
    lc++;

    return;
  }

  if (strlen(op) < 6) {
    message(MSG_ERROR, "movwi operand syntax error.");
    exit(1);
  }
  if (op[1] == '+') { m = 0; rp = 5; }
  if (op[1] == '-') { m = 1; rp = 5; }
  if (op[4] == '+') { m = 2; rp = 3; }
  if (op[4] == '-') { m = 3; rp = 3; }
  if ((m < 0) || (rp < 3)) {
    message(MSG_ERROR, "movwi operand syntax error.");
    exit(1);
  }

  r = (op[rp] - '0') << 2;

  if (pass) mmWrite(chipmem, lc, 0x0018 | r | m);
  lc++;
}


int do_opcode(void (*opcode_func)(const char *op),
              const char *opcode, const char *operands, const char *comment)
{
  int opstart = lc;
  int i;

  if (pass && mmHasData(chipmem, lc))
      message(MSG_WARNING, "Overwriting previously written location");

  (*opcode_func)(operands);

  // fprintf(stderr, "DEBUG: do_opcode: lc at start = %04x, at end = %04x\n", opstart, lc);
  if (lstp) {
    fprintf(lstp, "%04x", opstart);
    for (i=opstart; i<lc; i++) fprintf(lstp, " %04hx", mmRead(chipmem, i));
    fprintf(lstp, "\t\t%s %s\n", opcode, operands);
  }
  return (lc - opstart);
}


/*
 *  Checks 'opcode' for a valid opcode. If found, calls do_opcode
 *  and returns its result.  Otherwise returns -1 to indicate
 *  it wasn't a valid opcode (may still be a directive, though).
 */

int try_opcode(const char *opcode, const char *operands,
               const char *comment)
{
  switch (opcode[0]) {
    case 'a':	// 8 opcodes
      if (!strcmp(opcode, "addwf"))
          return do_opcode(do_addwf,  opcode, operands, comment);
      if (!strcmp(opcode, "addwfc"))
          return do_opcode(do_addwfc, opcode, operands, comment);
      if (!strcmp(opcode, "andwf"))
          return do_opcode(do_andwf,  opcode, operands, comment);
      if (!strcmp(opcode, "asrf"))
          return do_opcode(do_asrf,   opcode, operands, comment);
      if (!strcmp(opcode, "addlw"))
          return do_opcode(do_addlw,  opcode, operands, comment);
      if (!strcmp(opcode, "andlw"))
          return do_opcode(do_andlw,  opcode, operands, comment);
      if (!strcmp(opcode, "addfsr0"))
          return do_opcode(do_addfsr0, opcode, operands, comment);
      if (!strcmp(opcode, "addfsr1"))
          return do_opcode(do_addfsr1, opcode, operands, comment);
      break;
    case 'b':	// 6 opcodes
      if (!strcmp(opcode, "bcf"))
          return do_opcode(do_bcf,    opcode, operands, comment);
      if (!strcmp(opcode, "bsf"))
          return do_opcode(do_bsf,    opcode, operands, comment);
      if (!strcmp(opcode, "btfsc"))
          return do_opcode(do_btfsc,  opcode, operands, comment);
      if (!strcmp(opcode, "btfss"))
          return do_opcode(do_btfss,  opcode, operands, comment);
      if (!strcmp(opcode, "bra"))
          return do_opcode(do_bra,    opcode, operands, comment);
      if (!strcmp(opcode, "brw"))
          return do_opcode(do_brw,    opcode, operands, comment);
      break;
    case 'c':	// 8 opcodes
      if (!strcmp(opcode, "clrf"))
          return do_opcode(do_clrf,   opcode, operands, comment);
      if (!strcmp(opcode, "clrw"))
          return do_opcode(do_clrw,   opcode, operands, comment);
      if (!strcmp(opcode, "call"))
          return do_opcode(do_call,   opcode, operands, comment);
      if (!strcmp(opcode, "call_far"))
          return do_opcode(do_call_far,  opcode, operands, comment);
      if (!strcmp(opcode, "call_near"))
          return do_opcode(do_call_near, opcode, operands, comment);
      if (!strcmp(opcode, "comf"))
          return do_opcode(do_comf,   opcode, operands, comment);
      if (!strcmp(opcode, "callw"))
          return do_opcode(do_callw,  opcode, operands, comment);
      if (!strcmp(opcode, "clrwdt"))
          return do_opcode(do_clrwdt, opcode, operands, comment);
      break;
    case 'd':
      if (!strcmp(opcode, "decf"))
          return do_opcode(do_decf,   opcode, operands, comment);
      if (!strcmp(opcode, "decfsz"))
          return do_opcode(do_decfsz, opcode, operands, comment);
      break;
    case 'e':
    case 'f':
      break;
    case 'g':
      if (!strcmp(opcode, "goto"))
          return do_opcode(do_goto,   opcode, operands, comment);
      if (!strcmp(opcode, "goto_far"))
          return do_opcode(do_goto_far,  opcode, operands, comment);
      if (!strcmp(opcode, "goto_near"))
          return do_opcode(do_goto_near, opcode, operands, comment);
      break;
    case 'h':
      break;
    case 'i':
      if (!strcmp(opcode, "incf"))
          return do_opcode(do_incf,   opcode, operands, comment);
      if (!strcmp(opcode, "iorwf"))
          return do_opcode(do_iorwf,  opcode, operands, comment);
      if (!strcmp(opcode, "incfsz"))
          return do_opcode(do_incfsz, opcode, operands, comment);
      if (!strcmp(opcode, "iorlw"))
          return do_opcode(do_iorlw,  opcode, operands, comment);
      break;
    case 'j':
    case 'k':
      break;
    case 'l':
      if (!strcmp(opcode, "lslf"))
          return do_opcode(do_lslf,   opcode, operands, comment);
      if (!strcmp(opcode, "lsrf"))
          return do_opcode(do_lsrf,   opcode, operands, comment);
      break;
    case 'm':	// 7 opcodes
      if (!strcmp(opcode, "movf"))
          return do_opcode(do_movf,   opcode, operands, comment);
      if (!strcmp(opcode, "movwf"))
          return do_opcode(do_movwf,  opcode, operands, comment);
      if (!strcmp(opcode, "movlb"))
          return do_opcode(do_movlb,  opcode, operands, comment);
      if (!strcmp(opcode, "movlp"))
          return do_opcode(do_movlp,  opcode, operands, comment);
      if (!strcmp(opcode, "movlw"))
          return do_opcode(do_movlw,  opcode, operands, comment);
      if (!strcmp(opcode, "moviw"))
          return do_opcode(do_moviw,  opcode, operands, comment);
      if (!strcmp(opcode, "movwi"))
          return do_opcode(do_movwi,  opcode, operands, comment);
      break;
    case 'n':
      if (!strcmp(opcode, "nop"))
          return do_opcode(do_nop,    opcode, operands, comment);
      break;
    case 'o':
      if (!strcmp(opcode, "option"))
          return do_opcode(do_option, opcode, operands, comment);
      break;
    case 'p':
    case 'q':
      break;
    case 'r':	// 6 opcodes
      if (!strcmp(opcode, "return"))
          return do_opcode(do_return, opcode, operands, comment);
      if (!strcmp(opcode, "retlw"))
          return do_opcode(do_retlw,  opcode, operands, comment);
      if (!strcmp(opcode, "rlf"))
          return do_opcode(do_rlf,    opcode, operands, comment);
      if (!strcmp(opcode, "rrf"))
          return do_opcode(do_rrf,    opcode, operands, comment);
      if (!strcmp(opcode, "retfie"))
          return do_opcode(do_retfie, opcode, operands, comment);
      if (!strcmp(opcode, "reset"))
          return do_opcode(do_reset,  opcode, operands, comment);
      break;
    case 's':
      if (!strcmp(opcode, "subwf"))
          return do_opcode(do_subwf,  opcode, operands, comment);
      if (!strcmp(opcode, "subwfb"))
          return do_opcode(do_subwfb, opcode, operands, comment);
      if (!strcmp(opcode, "swapf"))
          return do_opcode(do_swapf,  opcode, operands, comment);
      if (!strcmp(opcode, "sublw"))
          return do_opcode(do_sublw,  opcode, operands, comment);
      if (!strcmp(opcode, "sleep"))
          return do_opcode(do_sleep,  opcode, operands, comment);
      break;
    case 't':
      if (!strcmp(opcode, "tris"))
          return do_opcode(do_tris,   opcode, operands, comment);
      break;
    case 'u':
    case 'v':
    case 'w':
      break;
    case 'x':
      if (!strcmp(opcode, "xorwf"))
          return do_opcode(do_xorwf,  opcode, operands, comment);
      if (!strcmp(opcode, "xorlw"))
          return do_opcode(do_xorlw,  opcode, operands, comment);
      break;
    default:
      break;
  }  // switch

  return -1;
}


